@@ -16,17 +16,19 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-struct Muxer_arg
-  {
-  int dictionary_size;
-  int match_len_limit;
-  int num_workers;
-  int num_slots;
-  int debug_level;
-  int infd;
-  int outfd;
-  };
+int compress( const int data_size, const int dictionary_size,
+              const int match_len_limit, const int num_workers,
+              const int num_slots, const int infd, const int outfd,
+              const int debug_level );
 
-void * muxer( void * arg );
+
+/*----------------------- Defined in main.cc -----------------------*/
+
+void show_error( const char * msg, const int errcode = 0, const bool help = false ) throw();
+int readblock( const int fd, uint8_t * buf, const int size ) throw();
+int writeblock( const int fd, const uint8_t * buf, const int size ) throw();
+
+
+void fatal();		// Terminate the process
 
 extern int verbosity;
