@@ -1,6 +1,6 @@
 #! /bin/sh
 # check script for Plzip - A parallel version of the lzip data compressor
-# Copyright (C) 2009 Antonio Diaz Diaz.
+# Copyright (C) 2009, 2010 Antonio Diaz Diaz.
 #
 # This script is free software: you have unlimited permission
 # to copy, distribute and modify it.
@@ -22,10 +22,11 @@ mkdir tmp
 echo -n "testing plzip..."
 cd "${objdir}"/tmp
 
-cat "${testdir}"/../COPYING > in || framework_failure
+cat "${testdir}"/test1 > in || framework_failure
+cat in in in in in > in5 || framework_failure
 fail=0
 
-"${LZIP}" -cd "${testdir}"/COPYING.lz > copy || fail=1
+"${LZIP}" -cd "${testdir}"/test1.lz > copy || fail=1
 cmp in copy || fail=1
 
 for i in s4096 1 2 3 4 5 6 7 8; do
@@ -46,7 +47,7 @@ for i in s4096 1 2 3 4 5 6 7 8; do
 done
 
 for i in s4096 1 2 3 4 5 6 7 8; do
-	"${LZIP}" -c -$i < in > out || fail=1
+	"${LZIP}" -$i < in > out || fail=1
 	"${LZIP}" -d < out > copy || fail=1
 	cmp in copy || fail=1
 	echo -n .
@@ -59,6 +60,13 @@ for i in s4096 1 2 3 4 5 6 7 8; do
 	echo -n .
 done
 
+for i in 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16; do
+	"${LZIP}" -s4096 -n$i < in5 > out5 || fail=1
+	"${LZIP}" -d < out5 > copy5 || fail=1
+	cmp in5 copy5 || fail=1
+	echo -n .
+done
+
 echo
 if [ ${fail} = 0 ] ; then
 	echo "tests completed successfully."
